/**
 * @file activity-handler.js
 * 
 * @brief This file contain activity handler implementation which handles
 *        activity on the sonos device or group. Each sonos has 3 states: -
 *        actualState, current active - lastState, previous - blinkState,
 *        special action aka doorbell or custome events
 * 
 * state switch occur usually occur if upnp command PLAY appear. other upnp
 * commmands: MUTE/UN-MUTE, SET/INC/DEC VOLUME, PAUSE/RESUME are skipped. on
 * UNDO or time elapsed active state ends and new state becomes active again: -
 * if blink state was active, then actual state is used - if actual state was
 * active, then last state is used blinkState is special and overrides actual
 * state which is put onhold , waiting for the blinkState to end.
 * 
 * Each state holds a restore data part (sonos device current state snapshot)
 * and are used to restore the sonos after state switch.
 * 
 * There are two possibilities of state switch: - actual and last state switch,
 * in this case sonos device state is stored - blink and actual state switch, in
 * this case sonos device state is stored - blink active, actual arrives, in
 * this case blink remains active and actual is put onhold, only activity id is
 * stored and returned back later when blink state is invalid.
 * 
 * OTHER: If user is directly changing the sonos, bypassing the dSS, then
 * different musics could be played New activity is ignored if the same activity
 * is already active. TODO If blink activity is active and another blink
 * different activity appears then Same activity can have a timeout flag and an
 * undo. The first one prevails, and others for the same activity are ignored.
 * 
 * If activity has dontResume flat set then it mean the previos state will not
 * be resumed.
 * 
 * 
 * @author: mobile2DS
 * 
 * @date: 12.1.2014 file created
 * 
 * @remark dated 13.05.2014 Enhanced with aka sonos group handling: a) each new
 *         group overwrites previous group, but restore data remains. Only on
 *         first group a restore data will be stored b) each new non-grouped
 *         activity overwrites group (TODO): - first undo/timeout group must be
 *         called to revert the sonos state - then call with the new non-grouped
 *         activity c) timeout/undo of the non-grouped activity when the group
 *         is active does not break the group, but ends silently: - put
 *         restore/activity info to the group state info for later - invalidate
 *         state for non-grouped activity and return ok, but dont restore
 *         anything yet d) timeout/undo of grouped activity: - invalidate group
 *         state and return restore/activity info for the next upnp action
 */

// Global constants
var ACTUAL_STATE_NAME = "actual_state";
var LAST_STATE_NAME = "last_state";
var BLINK_STATE_NAME = "blink_state";
var GROUP_STATE_NAME = "group_state";

var INVALID_ACTIVITY_ID = -1;

var UPDATE_COMPLETE_STATE = 0;
var UPDATE_STATE_DATA_ONLY = 1;
var UPDATE_RESTORE_DATA_ONLY = 2;

var GET_COMPLETE_STATE = 0;
var GET_STATE_DATA_ONLY = 1;
var GET_RESTORE_DATA_ONLY = 2;

var STATE_INVALID = 0;
var STATE_INIT = 1;
var STATE_VALID = 2;

var EVENT_TYPE_CALL_SCENE = 0;
var EVENT_TYPE_CUSTOM = 1;
// Fill in any additional event types

var DATE_TIME_INVALID = 0;

/**
 * @brief Creates more states (actual, last, ..) for given device or group in
 *        the property tree.
 * 
 * @param
 * 
 * @return
 * 
 */
function createOneState(strBasePath, nItemId, strStateName) {
	lInfo.logln("createOneState: " + strStateName);
	try {
		var sub_par_path = nItemId + '/' + strStateName;

		// 0 - has data from actual state, 1 - has data from blink state
		storeSubParameter(strBasePath, sub_par_path + '/name', strStateName);
		storeSubParameter(strBasePath, sub_par_path + '/valid', STATE_INVALID);
		storeSubParameter(strBasePath, sub_par_path + '/upnp_cmd', COMMAND_NONE);
		storeSubParameter(strBasePath, sub_par_path + '/event_id', 0);
		storeSubParameter(strBasePath, sub_par_path + '/event_type', 0);
		storeSubParameter(strBasePath, sub_par_path + '/can_resume', false);
		storeSubParameter(strBasePath, sub_par_path + '/curr_activity_id', 0);
		storeSubParameter(strBasePath, sub_par_path + '/end_date_time_activity', DATE_TIME_INVALID);

		storeSubParameter(strBasePath, sub_par_path + '/activity_id', INVALID_ACTIVITY_ID);

		storeSubParameter(strBasePath, sub_par_path + '/valid_restore', 0);
		storeSubParameter(strBasePath, sub_par_path + '/id', 0);
		storeSubParameter(strBasePath, sub_par_path + '/playerIP', 0);
		storeSubParameter(strBasePath, sub_par_path + '/activeAction', 0);
		storeSubParameter(strBasePath, sub_par_path + '/isRadio', 0);
		storeSubParameter(strBasePath, sub_par_path + '/volumeDef', 0);
		storeSubParameter(strBasePath, sub_par_path + '/volumeSlaves', "");
		storeSubParameter(strBasePath, sub_par_path + '/playMode', 0);
		storeSubParameter(strBasePath, sub_par_path + '/isMuted', 0);
		storeSubParameter(strBasePath, sub_par_path + '/trackListPos', 0);
		storeSubParameter(strBasePath, sub_par_path + '/trackPlayPos', 0);
		storeSubParameter(strBasePath, sub_par_path + '/volumeReset', 0);
		storeSubParameter(strBasePath, sub_par_path + '/playUrl', 0);
		storeSubParameter(strBasePath, sub_par_path + '/activeQueueID', 0);
	} catch (e) {
		lInfo.logln("Error by creating one state: " + e);
		for ( var i in e) {
			lInfo.logln(i + ': ' + e[i]);
		}
	}
}

/**
 * @brief Creates more states (actual, last, blink, ...) for given device or
 *        group in the property tree.
 * 
 * @param
 * 
 * @return
 * 
 */
function createStates(nId, nIsGroup) {
	var have_it = false;
	var item_id = 0;
	var base_path = '';

	lInfo.logln("createStates: " + nId);

	try {
		item_id = nId;
		have_it = true;
		base_path = 'sonos-players';

		if (have_it) {
			createOneState(base_path, item_id, 'actual_state');

			createOneState(base_path, item_id, 'last_state');

			createOneState(base_path, item_id, 'blink_state');

			createOneState(base_path, item_id, 'group_state');

			// Property.store();
		}
	} catch (e) {
		lInfo.logln("Error by adding states to the device/group: " + e);
		for ( var i in e) {
			lInfo.logln(i + ': ' + e[i]);
		}
	}
}

/**
 * @brief Get state data
 * 
 * @param
 * 
 * @return
 * 
 */
function getStateData(nId, nIsGroup, strStateName) {
	var have_it = false;
	var item_id = 0;
	var base_path = '';
	var state = new Object();

	item_id = nId;
	have_it = true;
	base_path = 'sonos-players';

	try {
		var found_it = false;
		if (have_it) {
			var stateNode = Property.getNode(base_path + '/' + item_id + '/' + strStateName);
			if (stateNode != null) {
				state.stateName = stateNode.getChild('name').getValue();
				state.valid = stateNode.getChild('valid').getValue();
				state.eventId = stateNode.getChild('event_id').getValue();
				state.upnpCmd = stateNode.getChild('upnp_cmd').getValue();
				state.eventType = stateNode.getChild('event_type').getValue();
				state.canResume = stateNode.getChild('can_resume').getValue();
				state.currActivityId = stateNode.getChild('curr_activity_id').getValue();
				state.activityId = stateNode.getChild('activity_id').getValue();
				state.activityEndDateTime = stateNode.getChild('end_date_time_activity').getValue();
				state.restoreValid = stateNode.getChild('valid_restore').getValue();

				state.playerID = stateNode.getChild('id').getValue();
				state.playerIP = stateNode.getChild('playerIP').getValue();
				state.activeAction = stateNode.getChild('activeAction').getValue();
				state.isRadio = stateNode.getChild('isRadio').getValue();
				state.volumeDef = stateNode.getChild('volumeDef').getValue();
				state.volumeSlaves = stateNode.getChild('volumeSlaves').getValue();
				state.playMode = stateNode.getChild('playMode').getValue();
				state.isMuted = 0;
				if (stateNode.getChild('isMuted') != null) {
					state.isMuted = stateNode.getChild('isMuted').getValue();
				}
				state.trackListPos = stateNode.getChild('trackListPos').getValue();
				state.trackPlayPos = stateNode.getChild('trackPlayPos').getValue();
				state.volumeReset = stateNode.getChild('volumeReset').getValue();
				state.playUrl = stateNode.getChild('playUrl').getValue();
				state.activeQueueID = stateNode.getChild('activeQueueID').getValue();

				found_it = true;
			}
		}
	} catch (e) {
		lInfo.logln("Error by get state data: " + e);
		for ( var i in e) {
			lInfo.logln(i + ': ' + e[i]);
		}
	}

	if (found_it) {
		return state;
	} else {
		return null;
	}
}

/**
 * @brief Get actual state
 * 
 * @param
 * 
 * @return
 * 
 */
function getActualState(nId, nIsGroup) {
	return getStateData(nId, nIsGroup, 'actual_state');
}

/**
 * @brief Get last state
 * 
 * @param
 * 
 * @return
 * 
 */
function getLastState(nId, nIsGroup) {
	return getStateData(nId, nIsGroup, 'last_state');
}

/**
 * @brief Get blink state
 * 
 * @param
 * 
 * @return
 * 
 */
function getBlinkState(nId, nIsGroup) {
	return getStateData(nId, nIsGroup, 'blink_state');
}

/**
 * @brief Get group state
 * 
 * @param
 * 
 * @return
 * 
 */
function getGroupState(nId, nIsGroup) {
	return getStateData(nId, nIsGroup, 'group_state');
}

/**
 * @brief Update state data
 * 
 * @param
 * 
 * @return
 * 
 */
function updateStateData(nId, nIsGroup, strStateName, oState, nUpdateWhat) {
	var have_it = false;
	var item_id = 0;
	var base_path = '';
	var updated = false;

	lInfo.logln("updateStateData: " + nId);

	try {
		item_id = nId;
		have_it = true;
		base_path = 'sonos-players';

		if (have_it) {

			lInfo.logln("updateStateData: will store changed data into the state");
			var stateNode = Property.getNode(base_path + '/' + item_id + '/' + strStateName);
			if (stateNode != null) {
				if (nUpdateWhat != UPDATE_RESTORE_DATA_ONLY) {
					storeSubParameter(base_path, item_id + '/' + strStateName + '/valid', oState.valid);
					storeSubParameter(base_path, item_id + '/' + strStateName + '/upnp_cmd', oState.upnpCmd);
					storeSubParameter(base_path, item_id + '/' + strStateName + '/event_id', oState.eventId);
					storeSubParameter(base_path, item_id + '/' + strStateName + '/event_type', oState.eventType);
					storeSubParameter(base_path, item_id + '/' + strStateName + '/can_resume', oState.canResume);
					storeSubParameter(base_path, item_id + '/' + strStateName + '/end_date_time_activity', oState.activityEndDateTime);
					storeSubParameter(base_path, item_id + '/' + strStateName + '/curr_activity_id', oState.currActivityId);
				}

				if (nUpdateWhat != UPDATE_STATE_DATA_ONLY) {
					storeSubParameter(base_path, item_id + '/' + strStateName + '/activity_id', oState.activityId);
					storeSubParameter(base_path, item_id + '/' + strStateName + '/valid_restore', oState.restoreValid);
					storeSubParameter(base_path, item_id + '/' + strStateName + '/id', nId);
					storeSubParameter(base_path, item_id + '/' + strStateName + '/playerIP', oState.playerIP);
					storeSubParameter(base_path, item_id + '/' + strStateName + '/activeAction', oState.activeAction);
					storeSubParameter(base_path, item_id + '/' + strStateName + '/isRadio', oState.isRadio);
					storeSubParameter(base_path, item_id + '/' + strStateName + '/volumeDef', oState.volumeDef);
					storeSubParameter(base_path, item_id + '/' + strStateName + '/volumeSlaves', oState.volumeSlaves);
					storeSubParameter(base_path, item_id + '/' + strStateName + '/playMode', oState.playMode);
					if (oState.isMuted != null) {
						storeSubParameter(base_path, item_id + '/' + strStateName + '/isMuted', oState.isMuted);
					}
					storeSubParameter(base_path, item_id + '/' + strStateName + '/trackListPos', oState.trackListPos);
					storeSubParameter(base_path, item_id + '/' + strStateName + '/trackPlayPos', oState.trackPlayPos);
					storeSubParameter(base_path, item_id + '/' + strStateName + '/volumeReset', oState.volumeReset);
					storeSubParameter(base_path, item_id + '/' + strStateName + '/playUrl', oState.playUrl);
					storeSubParameter(base_path, item_id + '/' + strStateName + '/activeQueueID', oState.activeQueueID);
				}

				lInfo.logln("State updated");

				updated = true;
			}
		}
	} catch (e) {
		lInfo.logln("Error by update-ing state: " + e);
		for ( var i in e) {
			lInfo.logln(i + ': ' + e[i]);
		}
	}

	return updated;
}

/**
 * @brief Update actual state data
 * 
 * @param
 * 
 * @return
 * 
 */
function updateActualState(nId, nIsGroup, oState, nUpdateWhat) {
	return updateStateData(nId, nIsGroup, 'actual_state', oState, nUpdateWhat);
}

/**
 * @brief Update last state data
 * 
 * @param
 * 
 * @return
 * 
 */
function updateLastState(nId, nIsGroup, oState, nUpdateWhat) {
	return updateStateData(nId, nIsGroup, 'last_state', oState, nUpdateWhat);
}

/**
 * @brief Update blink state data
 * 
 * @param
 * 
 * @return
 * 
 */
function updateBlinkState(nId, nIsGroup, oState, nUpdateWhat) {
	return updateStateData(nId, nIsGroup, 'blink_state', oState, nUpdateWhat);
}

/**
 * @brief Update group state data
 * 
 * @param
 * 
 * @return
 * 
 */
function updateGroupState(nId, nIsGroup, oState, nUpdateWhat) {
	return updateStateData(nId, nIsGroup, 'group_state', oState, nUpdateWhat);
}

/**
 * @brief Check if its blink activity aka doorbell
 * 
 * @param
 * 
 * @returne
 * 
 */
function isBlinkActivity(eventId, eventType) {
	if ((eventId == 73 /* doorbell */) && (eventType == EVENT_TYPE_CALL_SCENE)) {
		return true;
	}
	return false;
}

/**
 * @brief Setup restore data for the caller
 * 
 * @param
 * 
 * @return
 * 
 */
function setupRestoreData(refRestoreData, refState) {
	lInfo.logln("setupRestoreData");

	refRestoreData.id = refState.playerID;
	refRestoreData.playerIP = refState.playerIP;
	refRestoreData.activeAction = refState.activeAction;
	refRestoreData.isRadio = refState.isRadio;
	refRestoreData.volumeDef = refState.volumeDef;
	refRestoreData.volumeSlaves = refState.volumeSlaves;
	refRestoreData.playMode = refState.playMode;
	refRestoreData.isMuted = refState.isMuted;
	refRestoreData.trackListPos = refState.trackListPos;
	refRestoreData.trackPlayPos = refState.trackPlayPos;
	refRestoreData.volumeReset = refState.volumeReset;
	refRestoreData.playUrl = refState.playUrl;
	refRestoreData.activeQueueID = refState.activeQueueID;
}

/**
 * @brief Reset (invalidate) device/group states
 * 
 * @param
 * 
 * @return
 * 
 */
function resetStates(nId, nIsGroup) {
	lInfo.logln("resetStates for player " + nId);

	try {
		var base_path = 'sonos-players';

		var stateNode = Property.getNode('sonos-players' + '/' + nId);
		if (baseNode != null) {

			lInfo.logln("delete existing states for player " + nId);

			var oArrayNode = baseNode.getChildren();
			for ( var iIndex = 0; iIndex < oArrayNode.length; iIndex++) {
				var oChildNode = oArrayNode[iIndex];
				if (oChildNode.getChild('actual_state') != null) {
					baseNode.removeChild(oChildNode.getChild('actual_state').getValue());
				}

				if (oChildNode.getChild('last_state') != null) {
					baseNode.removeChild(oChildNode.getChild('last_state').getValue());
				}

				if (oChildNode.getChild('blink_state') != null) {
					baseNode.removeChild(oChildNode.getChild('blink_state').getValue());
				}

				if (oChildNode.getChild('group_state') != null) {
					baseNode.removeChild(oChildNode.getChild('group_state').getValue());
				}
			}
		}

		createStates(nId, nIsGroup);
	} catch (e) {
		lInfo.logln("Error resetting states: " + e);
		for ( var i in e) {
			lInfo.logln(i + ': ' + e[i]);
		}
	}
}

/**
 * @brief Store restore data into the proper state
 * 
 * @param
 * 
 * @return
 * 
 */
function storeRestartDataIntoState(refRestoreData, refState) {
	lInfo.logln("storeRestartDataIntoState");

	refState.playerID = refRestoreData.id;
	refState.playerIP = refRestoreData.playerIP;
	refState.activeAction = refRestoreData.activeAction;
	refState.isRadio = refRestoreData.isRadio;
	refState.volumeDef = refRestoreData.volumeDef;
	refState.volumeSlaves = refRestoreData.volumeSlaves;
	refState.playMode = refRestoreData.playMode;
	refState.isMuted = refRestoreData.isMuted;
	refState.trackListPos = refRestoreData.trackListPos;
	refState.trackPlayPos = refRestoreData.trackPlayPos;
	refState.volumeReset = refRestoreData.volumeReset;
	refState.playUrl = refRestoreData.playUrl;
	refState.activeQueueID = refRestoreData.activeQueueID;
}

/**
 * @brief Set the restart data of the actual state
 * 
 * @param restartData -
 *            restart data parameters
 * 
 * @return 1 - actual state updated, 0 - could not update the actual state
 * 
 */
function storeUpnpDeviceStatus(restartData, nIsGroup) {
	lInfo.logln("storeUpnpDeviceStatus: " + restartData.id);

	var blink_state = getBlinkState(restartData.id, 0);
	var group_state = getGroupState(restartData.id, 0);
	var result = false;

	lInfo.logln("storeUpnpDeviceStatus: event id " + restartData.sceneID);

	if (nIsGroup == 1) {
		group_state.valid = STATE_VALID;
		group_state.restoreValid = 1;

		group_state.activityId = INVALID_ACTIVITY_ID;

		storeRestartDataIntoState(restartData, group_state);

		if (updateGroupState(restartData.id, 0, group_state, UPDATE_COMPLETE_STATE)) {
			lInfo.logln("storeUpnpDeviceStatus: Added device data to the group state");
			result = true;
		}
	} else if (blink_state.valid == STATE_INIT) {
		blink_state.valid = STATE_VALID;
		blink_state.restoreValid = 1;

		blink_state.activityId = INVALID_ACTIVITY_ID;

		storeRestartDataIntoState(restartData, blink_state);

		if (updateBlinkState(restartData.id, 0, blink_state, UPDATE_COMPLETE_STATE)) {
			lInfo.logln("storeUpnpDeviceStatus: Added device data to the blink state");
			result = true;
		}
	} else {
		var actual_state = getActualState(restartData.id, 0);

		if ((actual_state.valid == STATE_INIT) && (actual_state.canResume)) {
			lInfo.logln("storeUpnpDeviceStatus: actual state can resume");
			actual_state.restoreValid = 1;
			actual_state.valid = STATE_VALID;

			actual_state.activityId = INVALID_ACTIVITY_ID;

			if ((blink_state.valid == STATE_VALID) && (blink_state.restoreValid == 1)) {
				lInfo.logln("storeUpnpDeviceStatus: blink state is active -> should not occur");
				actual_state.valid = STATE_INVALID;
				actual_state.restoreValid = 0;
				actual_state.activityId = INVALID_ACTIVITY_ID;

				updateActualState(nId, 0, actual_state, UPDATE_STATE_DATA_ONLY);
			} else {
				storeRestartDataIntoState(restartData, actual_state);

				if (updateActualState(restartData.id, 0, actual_state, UPDATE_COMPLETE_STATE)) {
					lInfo.logln("storeUpnpDeviceStatus: Added device data to the actual state");
					result = true;
				}
			}
		} else if (actual_state.valid == STATE_INIT) {
			lInfo.logln("storeUpnpDeviceStatus: actual state cannot resume");

			actual_state.valid = STATE_VALID;
			actual_state.restoreValid = 0;
			actual_state.activityId = INVALID_ACTIVITY_ID;

			updateActualState(nId, 0, actual_state, UPDATE_STATE_DATA_ONLY);
		}
	}

	return result;
}

/**
 * @brief Store only the activity for the next action to executed after blink
 *        state ends
 * 
 * @param
 * 
 * @return 1 - actual state updated, 0 - could not update the actual state
 * 
 * @remark Blink state is active and actual state is postponed
 * 
 */
function storeActivityForLater(nId, nIsGroup, nActivityId) {
	lInfo.logln("storeActivityForLater: " + nId);

	var actual_state = getActualState(nId, nIsGroup);
	var blink_state = getBlinkState(nId, nIsGroup);
	var group_state = getGroupState(nId, nIsGroup);
	var result = false;

	if (nIsGroup == 1) {

	} else if (actual_state.valid == STATE_INIT) {
		actual_state.valid = STATE_VALID;
		actual_state.restoreValid = 0;

		actual_state.activityId = INVALID_ACTIVITY_ID;

		if ((blink_state.valid == STATE_VALID) && (blink_state.restoreValid == 1)) {
			// as blink state must end and only then we can play, dont store
			// the restore state of the blink state, but copy the restore data
			// from the blink state and put this restore data into the blink
			// state
			// so blink state will later restore us

			lInfo.logln("storeActivityForLater: blink state is active");

			lInfo.logln("storeActivityForLater: copy actual state with blink restore data");
			updateActualState(nId, 0, blink_state, UPDATE_RESTORE_DATA_ONLY);
			lInfo.logln("storeActivityForLater: copied device data from blink to actual state");

			blink_state.restoreValid = 0;

			blink_state.activityId = nActivityId;

			updateBlinkState(nId, 0, blink_state, UPDATE_COMPLETE_STATE);

			result = true;
		} else {
			lInfo.logln("storeActivityForLater: blink state is not valid -> should not happen");

			actual_state.valid = STATE_INVALID;
			actual_state.restoreValid = 0;
			actual_state.activityId = INVALID_ACTIVITY_ID;
		}

		updateActualState(nId, 0, actual_state, UPDATE_STATE_DATA_ONLY);
	}

	return result;
}

/**
 * @brief Handle new start activity (only dS activity)
 * 
 * @param nId -
 *            sonos device id or group id
 * @param nIsGroup -
 *            1 if its a group, 0 - single sonos device
 * @param eventId -
 *            detected event id
 * @param eventType -
 *            different event type (callscene, uda, ...)
 * @param upnpCmd -
 *            upnp cmd
 * @param canResume -
 *            does this state resumes previous state - true , false - not
 *            (resume to previous state)
 * @param nCurrActivityId -
 *            activity id of the executed activity used for power restore
 * 
 * @return true - can continue executing upnp cmd, false - stop
 * 
 */
function handleNewActivity(nId, nIsGroup, eventId, eventType, upnpCmd, canResume, nCurrActivityId, nActivityEndDateTime) {
	lInfo.logln("handleNewActivity: " + nId);

	var result = new Object();

	result.executeUpnpCmd = false;
	result.storeCurrDevData = false;
	result.storeActivity = false;
	result.restoreData = null;

	try {
		// get states for this device/group
		var actual_state = getActualState(nId, nIsGroup);
		var blink_state = getBlinkState(nId, nIsGroup);
		var last_state = getLastState(nId, nIsGroup);
		//var group_state = getGroupState(nId, nIsGroup);

		if (nIsGroup == 1) {
			lInfo.logln("handleNewActivity: grouped activity");

			/*if (group_state.valid == STATE_VALID) {
				if ((group_state.eventId == eventId) && (group_state.eventType == eventType) && (group_state.upnpCmd == upnpCmd)) {
					lInfo.logln("handleNewActivity: overwrite group state with the same grouped activity");
				} else {
					lInfo.logln("handleNewActivity: different group state already valid -> overwrite");
				}
				// Use restore data from previous group state
				result.storeCurrDevData = false;
			} else {
				lInfo.logln("handleNewActivity: group state is invalid");

				group_state.valid = STATE_INIT;
				result.storeCurrDevData = true;
			}*/
			result.storeCurrDevData = true;
			
			result.executeUpnpCmd = true;

			/*group_state.eventId = eventId;
			group_state.upnpCmd = upnpCmd;
			group_state.eventType = eventType;
			group_state.canResume = canResume;
			group_state.currActivityId = nCurrActivityId;
			group_state.activityEndDateTime = nActivityEndDateTime;

			updateGroupState(nId, nIsGroup, group_state, UPDATE_STATE_DATA_ONLY);*/
		} else if (isBlinkActivity(eventId, eventType) || (eventType != EVENT_TYPE_CALL_SCENE)) {
			// A doorbell or custom event usually

			lInfo.logln("handleNewActivity: blink activity");

			if (blink_state.valid == STATE_VALID) {
				if ((blink_state.eventId == eventId) && (blink_state.eventType == eventType) && (blink_state.upnpCmd == upnpCmd)) {
					lInfo.logln("handleNewActivity: overwrite blink state with the same blink activity");
				} else {
					lInfo.logln("handleNewActivity: different blink state already valid -> overwrite");
				}
				// Use restore data from previous blink state
				result.storeCurrDevData = false;
			} else {
				lInfo.logln("handleNewActivity: blink state is valid");

				blink_state.valid = STATE_INIT;
				result.storeCurrDevData = true;
			}

			result.executeUpnpCmd = true;

			blink_state.eventId = eventId;
			blink_state.upnpCmd = upnpCmd;
			blink_state.eventType = eventType;
			blink_state.canResume = canResume;
			blink_state.currActivityId = nCurrActivityId;
			blink_state.activityEndDateTime = nActivityEndDateTime;

			/*if (group_state.valid == STATE_VALID) {
				lInfo.logln("handleNewActivity: blink state is overriding the grouping");
				if (result.storeCurrDevData == true) {
					// use restore data from the grouping, but only if previous
					// activity was not blink
					lInfo.logln("handleNewActivity: blink state is using restore data from the grouping");
					result.storeCurrDevData = false;
					updateBlinkState(nId, nIsGroup, group_state, UPDATE_RESTORE_DATA_ONLY);

					blink_state.valid = STATE_VALID;
				}
				group_state.valid = STATE_INVALID;
				updateGroupState(nId, nIsGroup, group_state, UPDATE_STATE_DATA_ONLY);
			}*/

			lInfo.logln("handleNewActivity: blink state is valid");
			updateBlinkState(nId, nIsGroup, blink_state, UPDATE_STATE_DATA_ONLY);
		} else if (upnpCmd == COMMAND_PLAY) {
			var same_activity_in_actual_state = 0;
			var same_activity_in_last_state = 0;

			if ((last_state.valid == STATE_VALID) && (last_state.eventId == eventId) && (last_state.eventType == eventType) && (last_state.upnpCmd == upnpCmd)) {
				lInfo.logln("handleNewActivity: new activity is the same in last state");

				same_activity_in_last_state = 1;
			}

			if ((actual_state.valid == STATE_VALID) && (actual_state.eventId == eventId) && (actual_state.evenType == eventType) && (actual_state.upnpCmd == upnpCmd)) {
				lInfo.logln("handleNewActivity: overwrite actual state with the same activity");

				same_activity_in_actual_state = 1;
			} else {
				lInfo.logln("handleNewActivity: last state become actual state and set new actual state");
			}

			if (same_activity_in_actual_state == 0) {
				// Last state is now actual state

				if (last_state.valid == STATE_INVALID) {
					lInfo.logln("handleNewActivity: last state not valid yet -> update with actual state restore data");
					updateLastState(nId, nIsGroup, actual_state, UPDATE_RESTORE_DATA_ONLY);
				}

				updateLastState(nId, nIsGroup, actual_state, UPDATE_STATE_DATA_ONLY);
			}

			actual_state.eventId = eventId;
			actual_state.upnpCmd = upnpCmd;
			actual_state.eventType = eventType;
			actual_state.canResume = canResume;
			actual_state.valid = STATE_INIT;
			actual_state.currActivityId = nCurrActivityId;
			actual_state.activityEndDateTime = nActivityEndDateTime;

			if (blink_state.valid == STATE_VALID) {
				lInfo.logln("handleNewActivity: must store only activity to be executed next");

				result.executeUpnpCmd = false;
				result.storeActivity = true;

				/*if (group_state.valid == STATE_VALID) {
					lInfo.logln("handleNewActivity: actual state is overriding the grouping, but put onhold because of the blink state");
					result.restoreData = new Object();

					setupRestoreData(result.restoreData, group_state);

					lInfo.logln("handleNewActivity: return restore data");

					group_state.valid = STATE_INVALID;
					updateGroupState(nId, nIsGroup, group_state, UPDATE_STATE_DATA_ONLY);
				}*/
			} else {
				lInfo.logln("handleNewActivity: must store also upnp device data if can resume");

				result.executeUpnpCmd = true;
				if (actual_state.canResume) {
					if (same_activity_in_actual_state == 0) {
						lInfo.logln("handleNewActivity: store restore data")

						result.storeCurrDevData = true;

						/*if (group_state.valid == STATE_VALID) {
							lInfo.logln("handleNewActivity: actual state is overriding the grouping");
							if (result.storeCurrDevData == true) {
								// use restore data from the grouping, but only
								// if previous activity was not blink
								lInfo.logln("handleNewActivity: actual state is using restore data from the grouping");
								result.storeCurrDevData = false;
								updateActualState(nId, nIsGroup, group_state, UPDATE_RESTORE_DATA_ONLY);

								actual_state.valid = STATE_VALID;
							}
							group_state.valid = STATE_INVALID;
							updateGroupState(nId, nIsGroup, group_state, UPDATE_STATE_DATA_ONLY);
						}*/
					} else {
						lInfo.logln("handleNewActivity: same activity which can resume");
						actual_state.valid = STATE_VALID;
					}
				} else {
					lInfo.logln("handleNewActivity: actual cant resume");
					actual_state.valid = STATE_VALID;
				}
			}

			updateActualState(nId, nIsGroup, actual_state, UPDATE_STATE_DATA_ONLY);
		} else {
			lInfo.logln("handleNewActivity: unknown action and not handled");
		}
	} catch (e) {
		lInfo.logln("Error handling new activity: " + e);
		for ( var i in e) {
			lInfo.logln(i + ': ' + e[i]);
		}
	}

	return result;
}

/**
 * @brief Stop the activity (dS preset or dS activity)
 * 
 * @param nId -
 *            sonos device id or group id
 * @param nIsGroup -
 *            1 if its a group, 0 - single sonos device
 * 
 * @return activity ID or player restart Data
 * 
 */
function handleStopActivity(nId, nIsGroup, eventId, eventType, isUndo) {
	lInfo.logln("handleStopActivity on event " + eventId + " type " + eventType + ", is undo " + isUndo);

	try {
		// get states for this device/group
		var actual_state = getActualState(nId, nIsGroup);
		var last_state = getLastState(nId, nIsGroup);
		var blink_state = getBlinkState(nId, nIsGroup);
		var group_state = getGroupState(nId, nIsGroup);

		var result = new Object();

		result.restoreData = null;
		result.activityId = INVALID_ACTIVITY_ID;
		result.performStop = false;

		var change_state = false;

		if ((nIsGroup == 1) && (group_state.eventType == eventType) && (group_state.eventId == eventId) && (group_state.valid == STATE_VALID)) {
			lInfo.logln("handleStopActivity in group state");

			group_state.valid = STATE_INVALID;
			group_state.nCurrActivityId = INVALID_ACTIVITY_ID;

			if (group_state.restoreValid == 1) {
				// return restart data to execute upnp action
				group_state.restoreValid = 0;
				if (group_state.canResume) {
					result.restoreData = new Object();

					setupRestoreData(result.restoreData, group_state);

					lInfo.logln("handleStopActivity return restore data from group state");
				}
			} else if (group_state.activityId != INVALID_ACTIVITY_ID) {
				result.activityId = group_state.activityId;

				group_state.activityId = INVALID_ACTIVITY_ID;

				lInfo.logln("handleStopActivity return activity id from group state");
			} else {
				lInfo.logln("handleStopActivity group state has no valid restore or activity");
			}

			updateGroupState(nId, nIsGroup, group_state, UPDATE_COMPLETE_STATE);
		} else if ((blink_state.valid == STATE_VALID) && (isBlinkActivity(eventId, eventType)) || ((blink_state.eventId == eventId) && (blink_state.eventType == eventType))) {
			lInfo.logln("handleStopActivity on blink state");

			blink_state.valid = STATE_INVALID;
			blink_state.nCurrActivityId = INVALID_ACTIVITY_ID;

			if (blink_state.restoreValid == 1) {
				// return restart data to execute upnp action

				if (group_state != null && group_state.valid == STATE_VALID) {
					updateGroupState(nId, nIsGroup, blink_state, UPDATE_RESTORE_DATA_ONLY);

					lInfo.logln("handleStopActivity copy restore data from the blink state to the group state");
				} else if (blink_state.canResume) {

					result.restoreData = new Object();

					setupRestoreData(result.restoreData, blink_state);

					lInfo.logln("handleStopActivity return restore data from blink state");
				}

				blink_state.restoreValid = 0;
			} else if (blink_state.activityId != INVALID_ACTIVITY_ID) {
				if (group_state.valid == STATE_VALID) {
					updateGroupState(nId, nIsGroup, blink_state, UPDATE_RESTORE_DATA_ONLY);

					lInfo.logln("handleStopActivity copy activity id from the blink state to the group state");
				} else {
					result.activityId = blink_state.activityId;
				}

				blink_state.activityId = INVALID_ACTIVITY_ID;

				lInfo.logln("handleStopActivity return activity id from blink state");
			} else {
				lInfo.logln("handleStopActivity blink state has no valid restore or activity");
			}

			updateBlinkState(nId, nIsGroup, blink_state, UPDATE_COMPLETE_STATE);
		} else if ((actual_state.eventId == eventId) && (actual_state.eventType == eventType) && (actual_state.valid == STATE_VALID)) {

			lInfo.logln("handleStopActivity: undo the actual state");
			actual_state.valid = STATE_INVALID;
			actual_state.nCurrActivityId = INVALID_ACTIVITY_ID;

			if (blink_state.valid == STATE_VALID) {
				lInfo.logln("handleStopActivity: undo actual state when blink state active");

				if (blink_state.activityId != INVALID_ACTIVITY_ID) {
					// blink state holds the activity for the actual state to be
					// executed
					// but as we got undo, must again set activity to invalid
					// and
					// get restore data from the actual state (copied from the
					// blink state
					// before)
					blink_state.activityId = INVALID_ACTIVITY_ID;
					if (blink_state.canResume) {
						lInfo.logln("handleStopActivity: copy blink state with actual restore data");
						updateBlinkState(nId, 0, actual_state, UPDATE_RESTORE_DATA_ONLY);
					} else {
						lInfo.logln("handleStopActivity: invalidate blink state activity id");
						updateBlinkState(nId, 0, blink_state, UPDATE_RESTORE_DATA_ONLY);
					}
				} else {
					lInfo.logln("handleStopActivity: actual started before blink and is ending when blink state still active");
					if (blink_state.canResume) {
						lInfo.logln("handleStopActivity: copy blink state with actual restore data");
						updateBlinkState(nId, 0, actual_state, UPDATE_RESTORE_DATA_ONLY);
					}
				}
			} else if (actual_state.restoreValid == 1) {
				// return restart data to execute upnp action
				if (actual_state.canResume) {
					if (group_state.valid == STATE_VALID) {
						updateGroupState(nId, nIsGroup, actual_state, UPDATE_RESTORE_DATA_ONLY);

						lInfo.logln("handleStopActivity copy restore data from the actual state to the group state");
					} else {
						result.restoreData = new Object();

						setupRestoreData(result.restoreData, actual_state);

						lInfo.logln("handleStopActivity: return restore data");
					}
				} else {
					lInfo.logln("handleStopActivity: resume not performed");
				}
				actual_state.restoreValid = 0;
			} else if (actual_state.activityId != INVALID_ACTIVITY_ID) {
				if (group_state.valid == STATE_VALID) {
					updateGroupState(nId, nIsGroup, actual_state, UPDATE_RESTORE_DATA_ONLY);

					lInfo.logln("handleStopActivity copy activity id from the actual state to the group state");
				} else {
					result.activityId = actual_state.activityId;
				}
				actual_state.activityId = INVALID_ACTIVITY_ID;

				lInfo.logln("handleStopActivity: return activity id");
			} else {
				lInfo.logln("handleStopActivity: actual state has no valid restore or activity -> perform stop");
				result.performStop = true;
			}

			if (!actual_state.canResume) {
				last_state.valid = STATE_INVALID;
			}

			updateActualState(nId, nIsGroup, last_state, UPDATE_COMPLETE_STATE);
			last_state.valid = STATE_INVALID;
			last_state.nCurrActivityId = INVALID_ACTIVITY_ID;
			updateLastState(nId, nIsGroup, last_state, UPDATE_STATE_DATA_ONLY);
		} else if ((last_state.eventId == eventId) && (last_state.eventType == eventType) && (last_state.valid == STATE_VALID)) {
			lInfo.logln("handleStopActivity: undo the last state");

			if (actual_state.valid == STATE_VALID) {
				if (blink_state.valid == STATE_VALID) {
					lInfo.logln("handleStopActivity: undo last state when blink state active");
				}
				lInfo.logln("handleStopActivity: move restore data only to the actual state");

				updateActualState(nId, nIsGroup, last_state, UPDATE_RESTORE_DATA_ONLY);
			}

			last_state.valid = STATE_INVALID;
			last_state.nCurrActivityId = INVALID_ACTIVITY_ID;

			updateLastState(nId, nIsGroup, last_state, UPDATE_STATE_DATA_ONLY);
		} else {
			lInfo.logln("handleStopActivity: none state valid -> nothing to do");
		}

	} catch (e) {
		lInfo.logln("Error handling new activity: " + e);
		for ( var i in e) {
			lInfo.logln(i + ': ' + e[i]);
		}
	}

	return result;
}

/**
 * @brief Handle new undo activity (dS preset or dS activity)
 * 
 * @param nId -
 *            sonos device id or group id
 * @param nIsGroup -
 *            1 if its a group, 0 - single sonos device
 * 
 * @return activity ID or restartData
 * 
 */
function handleUndoActivity(nId, nIsGroup, eventId) {
	lInfo.logln("handleUndoActivity " + nId);

	return handleStopActivity(nId, nIsGroup, eventId, EVENT_TYPE_CALL_SCENE, 1);
}

/**
 * @brief Handle stop sonos action (timed playlist ended playing)
 * 
 * @param nId -
 *            sonos device id or group id
 * @param nIsGroup -
 *            1 if its a group, 0 - single sonos device
 * 
 * @return activity ID or restartData
 * 
 */
function handleActivityTimeout(nId, nIsGroup, eventId, eventType) {
	// get states for this device/group
	lInfo.logln("handleActivityTimeout: " + nId);

	return handleStopActivity(nId, nIsGroup, eventId, eventType, 0);
}

/**
 * @brief Return if group state for this player is active
 * 
 * @param nId -
 *            sonos device id or group id
 * 
 * @return active state type: 1 group state active, 0 group state not active
 * 
 */
function isActivityGroupingActive(nId) {
	// get states for this device/group
	lInfo.logln("isActivityGroupingActive: " + nId);

	var group_active = 0;

	try {
		// get states for this device/group
		var group_state = getGroupState(nId, 0);

		if (group_state.valid == STATE_VALID) {
			group_active = 1;
		}
	} catch (e) {
		lInfo.logln("Error get is group active state: " + e);
		for ( var i in e) {
			lInfo.logln(i + ': ' + e[i]);
		}
	}

	return group_active;
}

/**
 * @brief Return current active activity
 * 
 * @param nId -
 *            sonos device id or group id
 * 
 * @return active state type: 1 group state active, 0 group state not active
 * 
 */
function getCurrentActiveActivity(nId) {
	// get states for this device/group
	lInfo.logln("getCurrentActiveActivity: " + nId);

	var result = new Object();

	result.nActiveActivityId = INVALID_ACTIVITY_ID;
	result.eventType = 0;
	result.eventId = 0;
	result.activityEndDateTime = '';

	try {
		// get states for this device/group
		var group_state = getGroupState(nId, 0);
		var actual_state = getActualState(nId, 0);
		var blink_state = getBlinkState(nId, 0);

		if (group_state.valid == STATE_VALID) {
			result.nActiveActivityId = group_state.currActivityId;
			result.eventType = group_state.eventType;
			result.eventId = group_state.eventId;
			result.activityEndDateTime = group_state.activityEndDateTime;
		} else if (blink_state.valid == STATE_VALID) {
			result.nActiveActivityId = blink_state.currActivityId;
			result.eventType = blink_state.eventType;
			result.eventId = blink_state.eventId;
			result.activityEndDateTime = blink_state.activityEndDateTime;
		} else if (actual_state.valid == STATE_VALID) {
			result.nActiveActivityId = actual_state.currActivityId;
			result.eventType = actual_state.eventType;
			result.eventId = actual_state.eventId;
			result.activityEndDateTime = actual_state.activityEndDateTime;
		}
	} catch (e) {
		lInfo.logln("Error get current active activity: " + e);
		for ( var i in e) {
			lInfo.logln(i + ': ' + e[i]);
		}
	}

	return result;
}
